use maud::{html, Markup, Render};
use orgize::Org;
use serde::{Serialize, Deserialize};
use std::{path::Path, ffi::OsStr};

#[derive(Debug, Serialize, Deserialize)]
pub struct PostMetadata {
    pub slug: String,
    pub title: String,
    pub date: String,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct PostContent {
    pub slug: String,
    pub title: String,
    pub date: String,
    pub html: String,
}

impl Render for PostMetadata {
    fn render(&self) -> Markup {
        html! {
            li { (self.date) " - " a href=(format!("/blog/{}", self.slug)) { (self.title) } }
        }
    }
}

// Render the actual blog post as HTML.
pub fn blog_post(fname: String, post: String) -> Result<PostContent, bool> {
    // Search through /posts directory and find the post with either .md or .org extension
    // If the post is not found, return 404

    let ext = Path::new(&fname).extension().and_then(OsStr::to_str).unwrap();
    let ext_name = format!(".{}", ext);
    let slug = fname.replace("posts/", "").replace( &ext_name, "");
    let mut html = "".to_owned();
    let mut date = "".to_owned();
    let mut title = "".to_owned();
    if ext == "md" {
        let (parsed, content) = frontmatter::parse_and_find_content(&post).unwrap();
        let metadata = parsed.unwrap();
        date = metadata["date"].as_str().unwrap().to_owned();
        title = metadata["title"].as_str().unwrap().to_owned();
        html = comrak::markdown_to_html(&content, &comrak::ComrakOptions::default());
    } else if ext == "org" {
        let mut writer = Vec::new();

        let parsed = Org::parse(&post);
        let keywords = parsed.keywords();
        // Get date and title from keywords iterator

        for keyword in keywords {
            if keyword.key == "date" {
                date = keyword.value.to_string();
            } else if keyword.key == "title" {
                title = keyword.value.to_string();
            }
        }
        parsed.write_html(&mut writer).unwrap();
        html = String::from_utf8(writer).unwrap();
    }
    Ok(PostContent {
        slug,
        title,
        date,
        html,
    })
}
