use shared::*;
use worker::*;
use maud::{html, Markup, Render, DOCTYPE};
use serde::Deserialize;
use include_dir::{include_dir, Dir};
use time::{self, macros::format_description, format_description::well_known::Rfc2822};

#[derive(Deserialize, Debug)]
struct Project {
    name: String,
    description: String,
    link: ProjectUrl,
    languages: Vec<String>,
}

#[derive(Deserialize, Debug)]
struct ProjectUrl {
    display_text: String,
    link: String,
}

#[derive(Deserialize, Debug)]
struct ProjectConfig {
    projects: Vec<Project>,
    experiments: Vec<Project>,
}

static DHALL_DIR: Dir<'_> = include_dir!("$CARGO_MANIFEST_DIR/src/dhall");

impl Render for Project {
    fn render(&self) -> Markup {
        html! {
            div .project {
                h4 { ( self.name ) " " }
                @for language in &self.languages {
                    span .language .(language.to_lowercase()) { ( language ) }
                }
                p { ( self.description ) " " a href=(self.link.link) { (self.link.display_text) }}
            }
        }
    }
}

#[event(fetch)]
async fn main(req: Request, env: Env, _ctx: Context) -> Result<Response> {
    let router = Router::new();
    router
        .get_async("/", |req, _ctx| async move {
            let cache = Cache::default();
            let key = req.url()?.to_string();
            if let Some(mut res) = cache.get(&key, false).await? {
                let headers = res.headers_mut();
                let _ = headers.set("cf-cache", "hit");
                return Ok(res)
            }

            let mut res = Response::from_html(homepage().await.into_string())?;
            let headers = res.headers_mut();
            let _ = headers.set("Cache-Control", "public, max-age=14400");
            cache.put(key, res.cloned()?).await?;
            Ok(res)
        })
        .get_async("/assets/styles.css", |req, _ctx| async move {
            let cache = Cache::default();
            let key = req.url()?.to_string();
            if let Some(mut res) = cache.get(&key, false).await? {
                let headers = res.headers_mut();
                let _ = headers.set("cf-cache", "hit");
                return Ok(res)
            }
            
            let css = include_str!("styles.css");
            let mut res = Response::ok(css)?;
            let headers = res.headers_mut();
            let _ = headers.set("Content-Type", "text/css");
            let _ = headers.set("Cache-Control", "public, max-age=14400");
            cache.put(key, res.cloned()?).await?;
            Ok(res)
        })
        .get_async("/blog", |req, ctx| async move {
            let cache = Cache::default();
            let key = req.url()?.to_string();
            if let Some(mut res) = cache.get(&key, false).await? {
                let headers = res.headers_mut();
                let _ = headers.set("cf-cache", "hit");
                return Ok(res);
            }
            
            let index = ctx.env.d1("INDEX")?;
            let statement = index.prepare("SELECT * FROM posts ORDER BY date DESC");
            let posts: Vec<PostMetadata> = statement.all().await?.results()?;

            let html = html! {
                (header("/blog"))
                    body {
                        main {
                            h1 { "Blog" }
                            a href = "/rss" { "RSS" }
                            ul {
                                @for post in posts {
                                    (post);
                                }
                            }
                        }
                    }
            };
            let mut res = Response::from_html(html.into_string())?;
            let headers = res.headers_mut();
            let _ = headers.set("Cache-Control", "public, max-age=14400");
            cache.put(key, res.cloned()?).await?;
            return Ok(res);
        })
        .get_async("/blog/:slug", |req, ctx| async move {
            let cache = Cache::default();
            let key = req.url()?.to_string();
            if let Some(mut res) = cache.get(&key, false).await? {
                let headers = res.headers_mut();
                let _ = headers.set("cf-cache", "hit");
                return Ok(res)
            }
            
            let index = ctx.env.d1("INDEX")?;
            let slug = ctx.param("slug").unwrap();
            let statement = index.prepare("SELECT * FROM posts WHERE slug=?1");
            let query = statement.bind(&[slug.into()])?;
            let post = query.first::<shared::PostContent>(None).await?;

            let p = post.unwrap();
            let html_maud = maud::PreEscaped(p.html);

            let html = html! {
                (header(p.title.as_str()))
                    body {
                        main {
                            h1 { (p.title) }
                            p { (p.date) }
                            (html_maud)
                        }
                    }
            };

            let mut res = Response::from_html(html.into_string())?;
            let headers = res.headers_mut();
            let _ = headers.set("Cache-Control", "public, max-age=14400");
            cache.put(key, res.cloned()?).await?;
            return Ok(res);            
        })
        .get_async("/rss", |req, ctx| async move {
            let cache = Cache::default();
            let key = req.url()?.to_string();
            if let Some(mut res) = cache.get(&key, false).await? {
                let headers = res.headers_mut();
                let _ = headers.set("cf-cache", "hit");
                return Ok(res)
            }

            let index = ctx.env.d1("INDEX")?;
            let statement = index.prepare("SELECT * FROM posts ORDER BY date DESC");
            let posts: Vec<PostMetadata> = statement.all().await?.results()?;
            let rss_posts: Vec<rss::Item> = posts.into_iter().map(|p| {
		        let date = format!("{} 00:00:00 +00:00:00", p.date);
		        let format = format_description!("[year]-[month]-[day] [hour]:[minute]:[second] [offset_hour sign:mandatory]:[offset_minute]:[offset_second]");
		        let parsed_date = match time::OffsetDateTime::parse(&date, &format) {
			        Ok(r) => r,
			        Err(_e) => { time::OffsetDateTime::now_utc() },
		        };
                let pub_date = match parsed_date.format(&Rfc2822) {
                  	Ok(r) => r,
			        Err(_e) => { "".to_owned() },
                };
		        rss::ItemBuilder::default()
			        .title(Some(p.title))
			        .link(Some(format!("https://gabrielsimmer.com/blog/{}", p.slug)))
			        .pub_date(Some(pub_date))
			        .build()
	        }).collect();
            let channel = rss::ChannelBuilder::default()
                .title("Gabriel Simmer's Blog".to_owned())
                .link("https://gabrielsimmer.com/blog".to_owned())
                .description("Gabriel Simmer's Blog Posts.".to_owned())
                .items(rss_posts)
                .build();

            let mut res = Response::ok(channel.to_string())?;
            let headers = res.headers_mut();
            let _ = headers.set("Cache-Control", "public, max-age=14400");
            let _ = headers.set("Content-Type", "application/rss+xml");

            cache.put(key, res.cloned()?).await?;
            Ok(res)
        })
        .run(req, env).await
}

async fn homepage() -> Markup {
    let projects: ProjectConfig = serde_dhall::from_str(DHALL_DIR.get_file("projects.dhall").unwrap().contents_utf8().unwrap())
        .parse()
        .unwrap();

    html! {
        (header("/"))
            body {
                main {
                    h1 { "Gabriel Simmer" }
                    h2 { "Infrastructure and DevOps" }
                    p { a href = "/blog" { "Blog" } " " a href = "https://floofy.tech/@arch" rel = "me" { "Fediverse" } }
                    h3 { "Projects" }
                    @for project in projects.projects {
                        (project)
                    }
                    h3 { "Experiments" }
                    @for project in projects.experiments {
                        (project)
                    }
                }
            }
    }
}

fn header(page_title: &str) -> Markup {
    html! {
        (DOCTYPE)
            meta charset="utf-8";
        meta name="viewport" content="width=device-width, initial-scale=1";
        title { (page_title) }
        link rel="stylesheet" href="/assets/styles.css";
        link rel="alternate" type="application/rss+xml" title="Gabriel Simmer's Blog" href="/rss";
    }
}
